from typing import Union, List
import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy import func
from tqdm import tqdm

from metacatalog.models import Entry
from metacatalog_search.models import TSIndex


def search(
        session: sa.orm.Session,
        search_string: str,
        limit: int = None,
        query: sa.orm.Query = None,
        as_result=False,
        return_iterator=False
    ):
    """
    Perform a full text search on the database.

    .. note:: 
        This method is part of the `metacatalog-search Extension <https://github.com/vforwater/metacatalog-search>`_
        You need to install and activate the extension, before you can use it.

        .. code-block:: bash
            pip install metacatalog-search
        
        >> from metacatalog import ext
        >> ext.activate_extension('search', 'metacatalog_search.extension', 'SearchExtension')

    Parameters
    ----------
    session : sqlalchemy.orm.Session
        Session connected to the database of metacatalog
    search_string : str
        The pattern to be searched. The search string can contain more
        than a single term and terms can be combined using logical expressions
        like or: ``'|'``, and: ``'&'`` or not: ``'!'``
    limit : int
        If given, the search results will be limited.

        .. note::
            Right now, the result is not yet sorted by relevance. Hence, 
            a limit will **not** necessarily return the most relevant 
            search results.

    query : sa.orm.Query
        Optionally, a base query can be passed, that may already contain
        filters. If not passed, it defaults to return a list of
        :class:`Entries <metacatalog.models.Entry>`
    as_result : bool
        If True, the result will be wrapped by a list of
        :class:`ImmutableResultSets <metcatalog.util.results.ImmutableResultSet>`.
    return_iterator : bool
        If True the query will be returned, instead of the results.
        Defaults to False.

    """
    # import here, due to circular imports
    from metacatalog_search.extension import SearchExtension
    from metacatalog import models
        
    # check if a base query was passed
    if query is None:
        query = session.query(Entry).join(TSIndex)
    else:
        query = query.join(TSIndex)
    
    # Add the full text search filter
    query = query.filter(TSIndex.tokens.match(f"'{search_string}'", postgresql_regconfig=SearchExtension.LANGUAGE))

    # TODO: here we could do a order by relevance somehow

    # add limit if needed
    if limit is not None:
        query = query.limit(limit)

    # return logic
    if return_iterator:
        return query
    else:
        if as_result:
            # import here due to circular imports
            from metacatalog.util.results import ImmutableResultSet
            return [ImmutableResultSet(entry) for entry in query.all()]
        return query.all()


def reindex_search(
    session: sa.orm.Session, 
    entries: Union[str, List[int], List[Entry]],
    attributes: Union[str, List[str]],
    if_exists = 'replace',
    verbose = False
    ):
    """
    Reindex the given list of :class:`Entries <metacatalog.models.Entry>`
    for full text search. The entries can be given as a List of ORM models
    as returned by the :func`find_entry <metacatalog.api.find_entry>` API
    or a list of thier ids. Finally, ``'all'`` or ``'missing'`` entries
    can be specified as a string literal.

    .. note:: 
        This method is part of the `metacatalog-search Extension <https://github.com/vforwater/metacatalog-search>`_
        You need to install and activate the extension, before you can use it.

        .. code-block:: bash
            pip install metacatalog-search
        
        >> from metacatalog import ext
        >> ext.activate_extension('search', 'metacatalog_search.extension', 'SearchExtension')

    Parameters
    ----------
    session : sqlalchemy.orm.Session
        Session connected to the database of metacatalog
    entries : list
        List of Entries, their ids or the literal ``'all'`` or 
        ``'missing'``. Be careful with literals as they might
        take some substantial amout of time.
    attributes : list
        List of attribute names that will be indexed. Instead of a list, 
        a string literal can be passed. With ``'default'``, a pre-defined list
        will be used. If ``'all'``, any string-based attribute will be used.
    if_exists : str
        If a Entry already was indexed, if will by default be replaced. 
        With ``'omit'``, the re-index will be skipped and the model is returned.
        If ``'raise'`` a :any:`AttributeError` will be raised.
    verbose : bool
        If True, a progressbar will be displayed. Defaults to False.

    """
    # check the entries attribute
    if entries == 'all':
        entry_stream = session.query(Entry)
    elif entries == 'missing':
        indexed_ids = [id for id, in session.query(TSIndex.entry_id)]
        entry_stream = session.query(Entry).filter(Entry.id.notin_(indexed_ids))
    elif isinstance(entries, sa.orm.query.Query):
        entry_stream = entries
    elif all([isinstance(_, int) for _ in entries]):
        entry_stream = session.query(Entry).filter(Entry.id.in_(entries))
    elif all([isinstance(_, Entry) for _ in entries]):
        entry_stream = entries
    else:
        raise AttributeError('entries not given as supported format.')
    
    # handle verbosity
    if verbose:
        if hasattr(entry_stream, 'count'):
            l = entry_stream.count()
        else:
            l = len(entry_stream)
        entry_stream = tqdm(entry_stream, total=l)
    
    # start the indexing
    # TODO: this could be run in parallel 
    for entry in entry_stream:
        entry.create_search_index(attributes=attributes, if_exists=if_exists)



import warnings
from sqlalchemy.exc import SAWarning
from metacatalog import api

from metacatalog_search import models


def _connect(connection='default'):
    """
    Connect to metacatalog by merging the declarative base
    and creating Foreign Key declarations.
    Returns the merged base
    """
    # only import when needed
    from metacatalog.db.base import Base

    # merge the declarative bases
    # error the warning and ignore then
    try:
        with warnings.catch_warnings():
            # This will raise the Warning instead of printing it
            warnings.simplefilter('error', category=SAWarning)
            models.merge_declarative_base(Base.metadata)
    except SAWarning:
        # do nothing here as
        pass

    return Base


def install(connection='default', verbose=False):
    # get the merged base
    Base  = _connect(connection=connection)

    # get a database connection
    session = api.connect_database(connection)

    # install
    Base.metadata.create_all(bind=session.bind)

#    # migrate default values
#    migrate(connection=connection)


class ManageCli:
    def install(self, connection='default'):
        install(connection=connection, verbose=True)


if __name__ == '__main__':
    import fire
    fire.Fire(ManageCli)


from typing import List


def expand_dict_to_str(obj) -> List[str]:
    """
    Expand the dictionary to str recursively
    """
    # get a dict representation or return as list of str
    if hasattr(obj, 'to_dict'):
        d = obj.to_dict().values()
    elif isinstance(obj, dict):
        d = obj.values()
    else:
        d = [obj]

    # expand
    out = []
    for value in d:
        # handle recursion
        if isinstance(value, dict):
            out.extend(expand_dict_to_str(value))
        if isinstance(value, (list, tuple)):
            for v in value:
                out.extend(expand_dict_to_str(v))
        
        # handle values
        if isinstance(value, (str, int, float)):
            out.extend([str(value)])
    
    return out        



import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import TSVECTOR, ARRAY


Base = declarative_base()


class TSIndex(Base):
    """
    """
    __tablename__ = 'ts_index'

    # columns
    id = sa.Column(sa.BigInteger, primary_key=True)
    attribute_names = sa.Column(ARRAY(sa.String(60)), nullable=False)
    tokens = sa.Column(TSVECTOR, nullable=False)


def merge_declarative_base(other: sa.MetaData):
    """
    Merge this declarative base with metacatalog declarative base
    to enable foreign keys and relationships between both classes.
    """
    from metacatalog import models

    # build missing columns
    #_connect_to_metacatalog()
    if not hasattr(TSIndex, 'entry_id'):
        TSIndex.entry_id  = sa.Column(sa.Integer, sa.ForeignKey(models.Entry.id), nullable=False)

    # add these tables to the other metadata
    TSIndex.__table__.to_metadata(other)

    # add relationships
    if not hasattr(TSIndex, 'entry'):
        TSIndex.entry = relationship(models.Entry)
    
    # add to models
    models.TSIndex = TSIndex



def _connect_to_metacatalog():
    """
    Call this method to creates missing columns and foreign keys.
    """
    # add the two foreign keys to Entry
    # we need to check if the columns are already there, as the extension might already
    # be loaded by metacatalog and the connection is already there
    if not hasattr(TSIndex, 'entry_id'):
        TSIndex.entry_id  = sa.Column(sa.Integer, sa.ForeignKey('entries.id'), nullable=False)



from typing import Union, List
from sqlalchemy.orm import object_session
from sqlalchemy import insert, update, delete, exists
from sqlalchemy import func
from sqlalchemy import event
from metacatalog.ext.base import MetacatalogExtensionInterface
from metacatalog.models import Entry
from metacatalog import api

from metacatalog_search import models
from metacatalog_search.api import search, reindex_search
from metacatalog_search.util import expand_dict_to_str


DEFAULT_ATTRIBUTES = ['title', 'abstract', 'comment', 'variable', 'authors', 'keywords', 'details']


def create_search_index(
    self: Entry,
    attributes: Union[str, List[str]] = 'default',
    if_exists = 'replace',
    commit=True,
    ):
    """
    (Re-)create full-text search vectors for this Entry. 
    Tokenized word stems are created for each attribute given and stored.
    Only indexed Entries can be found by the :func:`search API <metacatalog.api.search>`.

    .. note:: 
        This method is part of the `metacatalog-search Extension <https://github.com/vforwater/metacatalog-search>`_
        You need to install and activate the extension, before you can use it.

        .. code-block:: bash
            pip install metacatalog-search
        
        >> from metacatalog import ext
        >> ext.activate_extension('search', 'metacatalog_search.extension', 'SearchExtension')

    Parameters
    ----------
    attributes : list
        List of attribute names that will be indexed. Instead of a list, 
        a string literal can be passed. With ``'default'``, a pre-defined list
        will be used. If ``'all'``, any string-based attribute will be used.
    if_exists : str
        If a Entry already was indexed, if will by default be replaced. 
        With ``'omit'``, the re-index will be skipped and the model is returned.
        If ``'raise'`` a :any:`AttributeError` will be raised. 
    commit : bool
        If True, the created or updated full text search index will be directly
        added to thedatabase. If False, the method will return the respective
        SQL statement.

    """
    # get a session
    session = object_session(self)

    # here, we will need the Table object, not the ORM class
    table = models.TSIndex.__table__

    # check if there is already an index
    is_indexed = session.query(exists().where(models.TSIndex.entry_id == self.id)).scalar()

    # return or raise if already exists and not replace 
    if is_indexed and if_exists == 'raise':
        raise AttributeError(f'The Entry of ID={self.id} was already indexed.')
    elif is_indexed and if_exists == 'omit':
        return session.query(models.TSIndex).where(models.TSIndex.entry_id == self.id).one()

    # we will either insert or update
    if is_indexed:
        stmt = update(table).where(table.c.entry_id == self.id)
    else:
        stmt = insert(table)

    # handles the attributes list
    if attributes == 'default':
        attributes = DEFAULT_ATTRIBUTES
    if not isinstance(attributes, list) or any([not isinstance(_, str) for _ in attributes]):
        raise AttributeError(f"attributes has to be a list of names.")
    
    # collect all text chunks
    text_chunks = []
    attribute_names = []

    # collect the index information
    for name in attributes:
        if hasattr(self, name):
            text_chunks.extend(expand_dict_to_str(getattr(self, name)))
            attribute_names.append(name)
    
    # check if something was collected
    if len(text_chunks) == 0 or all([_ == '' for _ in text_chunks]):
        raise ValueError(f"This Entry cannot be indexed. No Strings found in {attributes}")
    
    # create the model
    stmt = stmt.values({
        'entry_id': self.id,
        'attribute_names': attribute_names,
        'tokens': func.to_tsvector(SearchExtension.LANGUAGE, ' '.join(text_chunks))
    })

    if not commit:
        return stmt

    # commit
    try:
        session.execute(stmt)
        session.commit()
    except Exception as e:
        session.rollback()
        raise e


def after_entry_insert(mapper, connection, target: Entry):
    """
    Call 
    """
    # TODO this is not yet working
    # get the correct statement
    # TODO how to handle the attributes list here?
#    stmt = target.create_search_index(commit=False)

    # execute
#    connection.execute(stmt)


def after_entry_update(mapper, connection, target: Entry):
    """
    Re-index the Entry after update
    """
    # TODO make this work again somehow
    # get the the update statement
#    stmt = target.create_search_index(commit=False)
#    connection.execute(stmt)


def before_entry_delete(mapper, connection, target: Entry):
    """
    Delete the search index before the Entry is deleted
    """
    stmt = delete(models.TSIndex).where(models.TSIndex.entry_id==target.id)
    connection.execute(stmt)


class SearchExtension(MetacatalogExtensionInterface):
    """
    Full Text Search extension. 
    This extension enables Metacatalog to search the database by
    PostgreSQL full-text search capabilities. A new table is installed
    to the database that indexes tokenized word stems for any specified
    text based attribute of the Entry.

    It adds a new instance method to Entry, to create a new search index
    for that specific Entry. Additionally, two API functions are added.
    A function to search and another function to re-index the whole table.

    Finally, new event listeners are added, that will re-create the search 
    index on Entry insert and update query. 

    """
    # TODO how to handle these settings better?
    LANGUAGE = 'english'
    AUTOINDEX = True

    @classmethod
    def init_extension(cls):
        """
        """
        # merge the declarative base from metacatalog
        from metacatalog.db.base import Base
        models.merge_declarative_base(Base.metadata)

        # add new instance method to Entry
        Entry.create_search_index = create_search_index

        # add new methods to API - check if this is working
        api.search = search
        api.reindex_search = reindex_search

        # TODO: add new event to index to Entry
        if cls.AUTOINDEX:
            event.listen(Entry, 'after_insert', after_entry_insert)
            event.listen(Entry, 'after_update', after_entry_update)

        #  add delete event to Entry
        event.listen(Entry, 'before_delete', before_entry_delete)
        


from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info


def requirements():
    with open('requirements.txt') as f:
        return f.read().split('\n')


def version():
    with open('VERSION') as f:
        return f.read().strip()


def readme():
    with open('README.md') as f:
        return f.read()

def install_and_migrate():
    try:
        from metacatalog_search import manage
    except ModuleNotFoundError:
        pass

    try:
        manage.install(verbose=True)
    except:
        pass


class PostDevelopCommand(develop):
    def run(self):
        print('DEVELOP RUNNING')
        develop.run(self)
        install_and_migrate()


class PostInstallCommand(install):
    def run(self):
        print('INSTALL RUNNNIG')
        install.run(self)
        install_and_migrate()


class PostEggInfoCommand(egg_info):
    def run(self):
        print('EGG_INFO RUNNING')
        egg_info.run(self)
        install_and_migrate()


setup(
    name='metacatalog_search',
    author='Mirko Mälicke',
    author_email='mirko.maelicke@kit.edu',
    license='GPL v3',
    install_requires=requirements(),
    version=version(),
    description='Full-Text Search extenstion for Metacatalog',
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    cmdclass={
        'develop': PostDevelopCommand,
        'install': PostInstallCommand,
        'egg_info': PostEggInfoCommand
    },
    include_package_data=True
)
